# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

import numpy as np

from ..property import PropertyMixin
from .entity import Entity
from ..value import Value, DataType
from . import util


class Property(Entity, PropertyMixin):

    def __init__(self, nixparent, h5dataset):
        super(Property, self).__init__(nixparent, h5dataset)
        self._h5dataset = self._h5group

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, dtype):
        util.check_entity_name(name)
        dtype = cls._make_h5_dtype(dtype)
        h5dataset = h5parent.create_dataset(name, shape=(0,), dtype=dtype)
        h5dataset.set_attr("name", name)
        h5dataset.set_attr("entity_id", util.create_id())
        newentity = cls(nixparent, h5dataset)
        newentity.force_created_at()
        newentity.force_updated_at()
        return newentity

    @property
    def name(self):
        return self._h5dataset.get_attr("name")

    @property
    def definition(self):
        return self._h5dataset.get_attr("definition")

    @definition.setter
    def definition(self, d):
        util.check_attr_type(d, str)
        self._h5dataset.set_attr("definition", d)

    @property
    def mapping(self):
        return self._h5dataset.get_attr("mapping")

    @mapping.setter
    def mapping(self, mapping):
        util.check_attr_type(mapping, str)
        self._h5dataset.set_attr("mapping", mapping)

    @property
    def unit(self):
        return self._h5dataset.get_attr("unit")

    @unit.setter
    def unit(self, u):
        util.check_attr_type(u, str)
        self._h5dataset.set_attr("unit", u)

    @property
    def values(self):
        dataset = self._h5dataset
        if not sum(dataset.shape):
            return tuple()
        data = np.empty(dataset.shape, dtype=dataset.dtype)
        self._h5dataset.read_data(data)

        def data_to_value(d):
            vitem = d["value"]
            if isinstance(vitem, bytes):
                vitem = vitem.decode()
            v = Value(vitem)
            v.uncertainty = d["uncertainty"]
            v.reference = d["reference"]
            v.filename = d["filename"]
            v.encoder = d["encoder"]
            v.checksum = d["checksum"]
            return v

        values = tuple(map(data_to_value, data))
        return values

    @values.setter
    def values(self, values):
        if not len(values):
            self.delete_values()
            return

        for v in values:
            util.check_attr_type(v, Value)

        self._h5dataset.shape = np.shape(values)

        dtype = self._h5dataset.dtype
        data = np.array([], dtype=dtype)
        for v in values:
            d = np.array((v.value, v.uncertainty, v.reference,
                          v.filename, v.encoder, v.checksum),
                         dtype=dtype)
            data = np.append(data, d)

        self._h5dataset.write_data(data)

    @property
    def data_type(self):
        dt = self._h5dataset.dtype[0].type
        if dt == util.vlen_str_dtype:
            return DataType.String
        else:
            return dt

    def delete_values(self):
        self._h5dataset.shape = (0,)

    @staticmethod
    def _make_h5_dtype(valuedtype):
        str_ = util.vlen_str_dtype
        common = [("uncertainty", float),
                  ("reference", str_),
                  ("filename",  str_),
                  ("encoder",   str_),
                  ("checksum",  str_)]

        if valuedtype == DataType.String:
            valuedtype = str_

        return np.dtype([("value", valuedtype)] + common)

    def __str__(self):
        return "{}: {{name = {}}}".format(
            type(self).__name__, self.name
        )

    def __repr__(self):
        return self.__str__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Copyright © 2014 German Neuroinformatics Node (G-Node)

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License. See
 LICENSE file in the root of the Project.

 Author: Jan Grewe <jan.grewe@g-node.org>

 This tutorial shows how to store a voltage trace and mark in this
 the occurence of action potentials.

 See https://github.com/G-node/nix/wiki for more information.

"""

import nixio as nix
import lif
import numpy as np
import matplotlib.pylab as plt


def fake_neuron():
    lif_model = lif.lif()
    t, v, spike_times = lif_model.run_const_stim(10000, 0.005)
    return t, v, spike_times


def plot_data(tag):
    data_array = tag.references[0]
    voltage = np.zeros(data_array.data.shape)
    data_array.data.read_direct(voltage)

    x_axis = data_array.dimensions[0]
    time = x_axis.axis(data_array.data.shape[0])

    spike_times = np.zeros(tag.positions.data_extent)
    tag.positions.data.read_direct(spike_times)

    plt.plot(time, voltage, color='dodgerblue', label=data_array.name)
    plt.scatter(spike_times, np.ones(spike_times.shape)*np.max(voltage), color='red', label=tag.name)
    plt.xlabel(x_axis.label + ((" [" + x_axis.unit + "]") if x_axis.unit else ""))
    plt.ylabel(data_array.label + ((" [" + data_array.unit + "]") if data_array.unit else ""))
    plt.title(data_array.name)
    plt.xlim(0, np.max(time))
    plt.ylim((1.2 * np.min(voltage), 1.2 * np.max(voltage)))
    plt.legend()
    plt.show()


if __name__ == '__main__':
    time, voltage, spike_times = fake_neuron()

    # create a new file overwriting any existing content
    file_name = 'spike_tagging.h5'
    file = nix.File.open(file_name, nix.FileMode.Overwrite)

    # create a 'Block' that represents a grouping object. Here, the recording session.
    # it gets a name and a type
    block = file.create_block("block name", "nix.session")

    # create a 'DataArray' to take the membrane voltage
    data = block.create_data_array("membrane voltage", "nix.regular_sampled.time_series", data=voltage)
    data.label = "membrane voltage"
    # add descriptors for time axis
    time_dim = data.append_sampled_dimension(time[1]-time[0])
    time_dim.label = "time"
    time_dim.unit = "s"

    # create the positions DataArray
    positions = block.create_data_array("times", "nix.positions", data=spike_times)
    positions.append_set_dimension() # these can be empty
    positions.append_set_dimension()

    # create a MultiTag
    multi_tag = block.create_multi_tag("spike times", "nix.events.spike_times", positions)
    multi_tag.references.append(data)

    # let's plot the data from the stored information
    plot_data(multi_tag)
    file.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Copyright © 2014 German Neuroinformatics Node (G-Node)

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License. See
 LICENSE file in the root of the Project.

 Author: Jan Grewe <jan.grewe@g-node.org>

 This tutorial shows how to store a voltage trace, mark in this the
 occurence of action poentials and also save for each spike the
 evoking stimulus as a feature.

 See https://github.com/G-node/nix/wiki for more information.

"""

import nixio as nix
import lif
import numpy as np
import scipy.signal as signal
import matplotlib.pylab as plt


def fake_neuron(stepsize=0.001, offset=.8, sta_offset=100):
    stimulus = np.random.randn(100000) * 2.5
    b, a = signal.butter(8, 0.25)
    stimulus = signal.filtfilt(b, a, stimulus)
    lif_model = lif.lif(stepsize=stepsize, offset=offset)
    time, v, spike_times = lif_model.run_stimulus(stimulus)
    snippets = np.zeros((len(spike_times), 2 * sta_offset))

    for i,t in enumerate(spike_times):
        index = round(t/stepsize)
        if index < sta_offset:
            snip = stimulus[0:index + sta_offset]
            snippets[i, -len(snip):] = snip
        elif (index + sta_offset) > len(stimulus):
            snip = stimulus[index-sta_offset:]
            snippets[i, 0:len(snip)] = snip
        else:
            snippets[i,:] = stimulus[index - sta_offset:index + sta_offset]
    return time, v, spike_times, snippets


def plot_data(tag):
    data_array = tag.references[0]
    voltage = np.zeros(data_array.data.shape)
    data_array.data.read_direct(voltage)

    x_axis = data_array.dimensions[0]
    time = x_axis.axis(data_array.data_extent[0])

    spike_times = tag.positions[:]

    feature_data_array = tag.features[0].data
    snippets = tag.features[0].data[:]
    single_snippet = tag.retrieve_feature_data(3, 0)[:]

    snippet_time_dim = feature_data_array.dimensions[1]
    snippet_time = snippet_time_dim.axis(feature_data_array.data_extent[1])

    response_axis = plt.subplot2grid((2, 2), (0, 0), rowspan=1, colspan=2)
    single_snippet_axis = plt.subplot2grid((2, 2), (1, 0), rowspan=1, colspan=1)
    average_snippet_axis = plt.subplot2grid((2, 2), (1, 1), rowspan=1, colspan=1)

    response_axis.plot(time, voltage, color='dodgerblue', label=data_array.name)
    response_axis.scatter(spike_times, np.ones(spike_times.shape)*np.max(voltage), color='red', label=tag.name)
    response_axis.set_xlabel(x_axis.label + ((" [" + x_axis.unit + "]") if x_axis.unit else ""))
    response_axis.set_ylabel(data_array.label + ((" [" + data_array.unit + "]") if data_array.unit else ""))
    response_axis.set_title(data_array.name)
    response_axis.set_xlim(0, np.max(time))
    response_axis.set_ylim((1.2 * np.min(voltage), 1.2 * np.max(voltage)))
    response_axis.legend()

    single_snippet_axis.plot(snippet_time, single_snippet.T, color="red", label=("snippet No 4"))
    single_snippet_axis.set_xlabel(snippet_time_dim.label + ((" [" + snippet_time_dim.unit + "]") if snippet_time_dim.unit else ""))
    single_snippet_axis.set_ylabel(feature_data_array.label + ((" [" + feature_data_array.unit + "]") if feature_data_array.unit else ""))
    single_snippet_axis.set_title("single stimulus snippet")
    single_snippet_axis.set_xlim(np.min(snippet_time), np.max(snippet_time))
    single_snippet_axis.set_ylim((1.2 * np.min(snippets[3,:]), 1.2 * np.max(snippets[3,:])))
    single_snippet_axis.legend()

    mean_snippet = np.mean(snippets, axis=0)
    std_snippet = np.std(snippets, axis=0)
    average_snippet_axis.fill_between(snippet_time, mean_snippet+std_snippet, mean_snippet-std_snippet, color="red", alpha=0.5)
    average_snippet_axis.plot(snippet_time, mean_snippet, color="red", label=(feature_data_array.name + str(4)))
    average_snippet_axis.set_xlabel(snippet_time_dim.label + ((" [" + snippet_time_dim.unit + "]") if snippet_time_dim.unit else ""))
    average_snippet_axis.set_ylabel(feature_data_array.label + ((" [" + feature_data_array.unit + "]") if feature_data_array.unit else ""))
    average_snippet_axis.set_title("spike-triggered average")
    average_snippet_axis.set_xlim(np.min(snippet_time), np.max(snippet_time))
    average_snippet_axis.set_ylim((1.2 * np.min(mean_snippet-std_snippet), 1.2 * np.max(mean_snippet+std_snippet)))

    plt.subplots_adjust(left=0.15, top=0.875, bottom=0.1, right=0.98, hspace=0.35, wspace=0.25)
    plt.show()


if __name__ == '__main__':
    stepsize = 0.0001 # s
    sta_offset = 100 # samples
    time, voltage, spike_times, sts = fake_neuron(stepsize=0.0001, sta_offset=sta_offset)

    # create a new file overwriting any existing content
    file_name = 'spike_features.h5'
    file = nix.File.open(file_name, nix.FileMode.Overwrite)

    # create a 'Block' that represents a grouping object. Here, the recording session.
    # it gets a name and a type
    block = file.create_block("block name", "nix.session")

    # create a 'DataArray' to take the membrane voltage
    data = block.create_data_array("membrane voltage", "nix.regular_sampled.time_series", data=voltage)
    data.label = "membrane voltage"
    # add descriptors for the time axis
    time_dim = data.append_sampled_dimension(stepsize)
    time_dim.label = "time"
    time_dim.unit = "s"

    # create the positions DataArray
    positions = block.create_data_array("times", "nix.positions", data=spike_times)
    positions.append_set_dimension() # these can be empty
    positions.append_set_dimension()

    # create a MultiTag
    multi_tag = block.create_multi_tag("spike times", "nix.events.spike_times", positions)
    multi_tag.references.append(data)

    # save stimulus snippets in a DataArray
    snippets = block.create_data_array("spike triggered stimulus", "nix.regular_sampled.multiple_series", data=sts)
    snippets.label = "stimulus"
    snippets.unit = "nA"
    set_dim = snippets.append_set_dimension()
    # add a descriptor for the time axis
    dim = snippets.append_sampled_dimension(stepsize)
    dim.unit = "s"
    dim.label = "time"
    dim.offset = -sta_offset * stepsize

    # set snippets as an indexed feature of the multi_tag
    multi_tag.create_feature(snippets, nix.LinkType.Indexed)

    # let's plot the data from the stored information
    plot_data(multi_tag)
    file.close()



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

from .entity_with_metadata import EntityWithMetadata
from ..source import SourceMixin
from . import exceptions
from . import util


class Source(EntityWithMetadata, SourceMixin):

    def __init__(self, nixparent, h5group):
        super(Source, self).__init__(nixparent, h5group)
        # TODO: Validate Source container

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        newentity = super(Source, cls)._create_new(nixparent, h5parent,
                                                   name, type_)
        return newentity

    # Source
    def create_source(self, name, type_):
        """
        Create a new source as a child of the current Source.

        :param name: The name of the source to create.
        :type name: str
        :param type_: The type of the source.
        :type type_: str

        :returns: The newly created source.
        :rtype: Source
        """
        util.check_entity_name_and_type(name, type_)
        sources = self._h5group.open_group("sources", True)
        if name in sources:
            raise exceptions.DuplicateName("create_source")
        src = Source._create_new(self, sources, name, type_)
        return src

    # Source
    def _get_source_by_id(self, id_or_name):
        sources = self._h5group.open_group("sources")
        return Source(self, sources.get_by_id_or_name(id_or_name))

    def _get_source_by_pos(self, pos):
        sources = self._h5group.open_group("sources")
        return Source(self, sources.get_by_pos(pos))

    def _delete_source_by_id(self, id_):
        sources = self._h5group.open_group("sources")
        sources.delete(id_)

    def _source_count(self):
        sources = self._h5group.open_group("sources")
        return len(sources)

    @property
    def referring_objects(self):
        objs = []
        objs.extend(self.referring_data_arrays)
        objs.extend(self.referring_tags)
        objs.extend(self.referring_multi_tags)
        return objs

    @property
    def referring_data_arrays(self):
        return [da for da in self._parent.data_arrays if self in da.sources]

    @property
    def referring_tags(self):
        return [tg for tg in self._parent.tags if self in tg.sources]

    @property
    def referring_multi_tags(self):
        return [mt for mt in self._parent.multi_tags if self in mt.sources]



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function)

try:
    basestring = basestring
except NameError:  # 'basestring' is undefined, must be Python 3
    basestring = (str, bytes)


class ProxyList(object):
    """
    Object that turns getters, setters etc of a NIX entity into
    something that acts like a mix between a list and a map.
    """

    def __init__(self, obj, counter, getter, index_getter, deleter):
        self.__counter = getattr(obj, counter)
        self.__getter = getattr(obj, getter)
        self.__index_getter = getattr(obj, index_getter)
        self.__deleter = getattr(obj, deleter)

    def __len__(self):
        return self.__counter()

    def __getitem__(self, key):
        if isinstance(key, basestring):
            val = None
            try:
                val = self.__getter(key)
            except:
                pass

            if val is None:
                raise KeyError("The given key does not exist: " + key)

            return val
        elif isinstance(key, int):
            count = self.__counter()

            if key < 0:
                key = count + key

            if key >= count or key < 0:
                raise KeyError("Index out of bounds: {}".format(key))

            return self.__index_getter(key)
        else:
            raise TypeError(
                "The key must be an int or a string but was: {}".format(
                    type(key)
                )
            )

    def __delitem__(self, key):
        if hasattr(key, "id"):
            key = key.id

        elem = self.__getitem__(key)
        self.__deleter(elem.id)

    def __iter__(self):
        for i in range(0, self.__counter()):
            yield self.__index_getter(i)

    def items(self):
        for i in range(0, self.__counter()):
            elem = self.__index_getter(i)
            yield (elem.id, elem)

    def __contains__(self, key):
        if hasattr(key, "id"):
            key = key.id

        if isinstance(key, basestring):
            try:
                elem = self.__getter(key)
                return elem is not None
            except:
                return False
        else:
            return False

    def __str__(self):
        str_list = [str(e) for e in list(self)]
        return "[" + ", ".join(str_list) + "]"

    def __repr__(self):
        return str(self)


class RefProxyList(ProxyList):

    def __init__(self, obj, counter, getter, index_getter, deleter, appender):
        super(RefProxyList, self).__init__(obj, counter, getter,
                                           index_getter, deleter)
        self.__appender = getattr(obj, appender)

    def append(self, key):
        if hasattr(key, "id"):
            key = key.id

        if isinstance(key, basestring):
            self.__appender(key)
        else:
            raise TypeError("The only id strings or entities can be appended")

    def extend(self, keys):
        if hasattr(keys, '__iter__'):
            [self.append(key) for key in keys]
        else:
            self.append(keys)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Copyright © 2014 German Neuroinformatics Node (G-Node)

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License. See
 LICENSE file in the root of the Project.

 Author: Jan Grewe <jan.grewe@g-node.org>

 This tutorial shows how to store a voltage trace and mark in this the
 interval in which a stimulus was presented. The stimulus itself is
 linked to this interval as an untagged Feature.

 See https://github.com/G-node/nix/wiki for more information.

"""

import nixio as nix
import lif
import numpy as np
import scipy.signal as signal
import matplotlib.pylab as plt


def fake_neuron(stepsize=0.001, offset=.8):
    stimulus = np.random.randn(80000) * 2.5
    b, a = signal.butter(8, 0.125)
    stimulus = signal.filtfilt(b, a, stimulus)
    s = np.hstack((np.zeros(10000), stimulus, np.zeros(10000)))
    lif_model = lif.lif(stepsize=stepsize, offset=offset)
    time, v, spike_times = lif_model.run_stimulus(s)
    stimulus_onset = 10000*stepsize
    stimulus_duration = len(stimulus)*stepsize
    return time, v, stimulus, stimulus_onset, stimulus_duration


def plot_data(tag):
    data_array = tag.references[0]
    voltage = data_array[:]

    x_axis = data_array.dimensions[0]
    time = x_axis.axis(data_array.data_extent[0])

    stimulus_onset = tag.position
    stimulus_duration = tag.extent

    stimulus = tag.retrieve_feature_data(0)
    stimulus_array = tag.features[0].data

    stim_time_dim = stimulus_array.dimensions[0]
    stimulus_time = stim_time_dim.axis(stimulus_array.data_extent[0])

    response_axis = plt.subplot2grid((2, 2), (0, 0), rowspan=1, colspan=2)
    response_axis.tick_params(direction='out')
    response_axis.spines['top'].set_color('none')
    response_axis.spines['right'].set_color('none')
    response_axis.xaxis.set_ticks_position('bottom')
    response_axis.yaxis.set_ticks_position('left')

    stimulus_axis = plt.subplot2grid((2, 2), (1, 0), rowspan=1, colspan=2)
    stimulus_axis.tick_params(direction='out')
    stimulus_axis.spines['top'].set_color('none')
    stimulus_axis.spines['right'].set_color('none')
    stimulus_axis.xaxis.set_ticks_position('bottom')
    stimulus_axis.yaxis.set_ticks_position('left')

    response_axis.plot(time, voltage, color='dodgerblue', label=data_array.name, zorder=1)
    response_axis.set_xlabel(x_axis.label + ((" [" + x_axis.unit + "]") if x_axis.unit else ""))
    response_axis.set_ylabel(data_array.label + ((" [" + data_array.unit + "]") if data_array.unit else ""))
    response_axis.set_xlim(0, np.max(time))
    response_axis.set_ylim((1.2 * np.min(voltage), 1.2 * np.max(voltage)))
    response_axis.barh(1.2 * np.min(voltage), stim_duration, (1.2*np.max(voltage)) - (1.2*np.min(voltage)),
                       stim_onset, color='silver', alpha=0.25, zorder=0, label="stimulus epoch")
    response_axis.legend()

    stimulus_axis.plot(stimulus_time, stimulus, color="slategray", label="stimulus")
    stimulus_axis.set_xlabel(stim_time_dim.label + ((" [" + stim_time_dim.unit + "]") if stim_time_dim.unit else ""))
    stimulus_axis.set_ylabel(stimulus_array.label + ((" [" + stimulus_array.unit + "]") if stimulus_array.unit else ""))
    stimulus_axis.set_xlim(np.min(stimulus_time), np.max(stimulus_time))
    stimulus_axis.set_ylim(1.2 * np.min(stimulus), 1.2 * np.max(stimulus))
    stimulus_axis.legend()

    plt.subplots_adjust(left=0.15, top=0.875, bottom=0.1, right=0.98, hspace=0.45, wspace=0.25)
    plt.show()


if __name__ == '__main__':
    stepsize = 0.0001 # s
    time, voltage, stimulus, stim_onset, stim_duration = fake_neuron(stepsize=0.0001)

    # create a new file overwriting any existing content
    file_name = 'untagged_feature.h5'
    file = nix.File.open(file_name, nix.FileMode.Overwrite)

    # create a 'Block' that represents a grouping object. Here, the recording session.
    # it gets a name and a type
    block = file.create_block("block name", "nix.session")

    # create a 'DataArray' to take the membrane voltage
    data = block.create_data_array("membrane voltage", "nix.regular_sampled.time_series", data=voltage)
    data.label = "membrane voltage"
    time_dim = data.append_sampled_dimension(stepsize)
    time_dim.label = "time"
    time_dim.unit = "s"

    # create a stimulus DataArray
    stim = block.create_data_array("stimulus", "nix.regular_sampled.time_series", data=stimulus)
    stim.label = "stimulus current"
    stim.unit = "nA"
    time_dim = stim.append_sampled_dimension(stepsize)
    time_dim.label = "time"
    time_dim.unit = "s"

    # create a Tag
    tag = block.create_tag("stimulus presentation", "nix.epoch.stimulus_presentation", [stim_onset])
    tag.extent = [stim_duration]
    tag.references.append(data)

    # set stimulus as untagged feature of the tag
    tag.create_feature(stim, nix.LinkType.Untagged)

    # let's plot the data from the stored information
    plot_data(tag)
    file.close()



class DuplicateName(Exception):

    def __init__(self, caller, *args, **kwargs):
        self.message = ("Duplicate name - "
                        "names have to be unique for a given "
                        "entity type & parent. ({})").format(caller)
        super(DuplicateName, self).__init__(self.message, *args, **kwargs)


class UninitializedEntity(Exception):

    def __init__(self, *args, **kwargs):
        self.message = "The Entity being accessed is uninitialized or empty."
        super(UninitializedEntity, self).__init__(self.message,
                                                  *args, **kwargs)


class InvalidUnit(Exception):

    def __init__(self, what, where):
        self.message = "InvalidUnit: {} evoked at: {}".format(what, where)
        super(InvalidUnit, self).__init__(self.message)


class InvalidAttrType(TypeError):

    def __init__(self, type_, value):
        self.message = ("Attribute requires type {} but {} "
                        "was provided.".format(type_, type(value)))
        super(InvalidAttrType, self).__init__(self.message)


class InvalidEntity(Exception):

    def __init__(self, *args, **kwargs):
        self.message = "Invalid entity found in HDF5 file."
        super(InvalidEntity, self).__init__(self.message, *args, **kwargs)


class OutOfBounds(IndexError):

    def __init__(self, message, index=None):
        self.message = message
        if index is not None:
            self.message += " [at index: {}]".format(index)
        super(OutOfBounds, self).__init__(self.message)


class IncompatibleDimensions(ValueError):

    def __init__(self, what, where):
        self.message = "IncompatibleDimensions: {} evoked at: {})".format(
            what, where
        )
        super(IncompatibleDimensions, self).__init__(self.message)


class InvalidFile(Exception):

    def __init__(self):
        self.message = "Invalid file - file is not a nix file."
        super(InvalidFile, self).__init__(self.message)



# -*- coding: utf-8 -*-
"""
 Copyright © 2014 German Neuroinformatics Node (G-Node)

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License. See
 LICENSE file in the root of the Project.

 Author: Jan Grewe <jan.grewe@g-node.org>
"""
import numpy as np


class LIF(object):

    def __init__(self, stepsize=0.0001, offset=1.6, tau_m=0.025, tau_a=0.02, da=0.0, D=3.5):
        self.stepsize = stepsize  # simulation stepsize [s]
        self.offset = offset  # offset curent [nA]
        self.tau_m = tau_m  # membrane time_constant [s]
        self.tau_a = tau_a  # adaptation time_constant [s]
        self.da = da  # increment in adaptation current [nA]
        self.D = D  # noise intensity
        self.v_threshold = 1.0  # spiking threshold
        self.v_reset = 0.0  # reset voltage after spiking
        self.i_a = 0.0  # current adaptation current
        self.v = self.v_reset  # current membrane voltage
        self.t = 0.0  # current time [s]
        self.membrane_voltage = []
        self.spike_times = []

    def _reset(self):
        self.i_a = 0.0
        self.v = self.v_reset
        self.t = 0.0
        self.membrane_voltage = []
        self.spike_times = []

    def _lif(self, stimulus, noise):
        """
        euler solution of the membrane equation with adaptation current and noise
        """
        self.i_a -= self.i_a - self.stepsize/self.tau_a * (self.i_a)
        self.v += self.stepsize * ( -self.v + stimulus + noise + self.offset - self.i_a)/self.tau_m;
        self.membrane_voltage.append(self.v)

    def _next(self, stimulus):
        """
        working horse which delegates to the euler and gets the spike times
        """
        noise = self.D * (float(np.random.randn() % 10000) - 5000.0)/10000
        self._lif(stimulus, noise)
        self.t += self.stepsize
        if self.v > self.v_threshold and len(self.membrane_voltage) > 1:
            self.v = self.v_reset
            self.membrane_voltage[len(self.membrane_voltage)-1] = 2.0
            self.spike_times.append(self.t)
            self.i_a += self.da

    def run_const_stim(self, steps, stimulus):
        """
        lif simulation with constant stimulus.
        """
        self._reset()
        for i in range(steps):
            self._next(stimulus);
        time = np.arange(len(self.membrane_voltage))*self.stepsize
        return time, np.array(self.membrane_voltage), np.array(self.spike_times)

    def run_stimulus(self, stimulus):
        """
        lif simulation with a predefined stimulus trace.
        """
        self._reset()
        for s in stimulus:
            self._next(s);
        time = np.arange(len(self.membrane_voltage))*self.stepsize
        return time, np.array(self.membrane_voltage), np.array(self.spike_times)

    def __str__(self):
        out = '\n'.join(["stepsize: \t" + str(self.stepsize),
                         "offset:\t\t" + str(self.offset),
                         "tau_m:\t\t" + str(self.tau_m),
                         "tau_a:\t\t" + str(self.tau_a),
                         "da:\t\t" + str(self.da),
                         "D:\t\t" + str(self.D),
                         "v_threshold:\t" + str(self.v_threshold),
                         "v_reset:\t" + str(self.v_reset)])
        return out

    def __repr__(self):
        return self.__str__()



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function)

import nixio.util.find as finders
from nixio.util.proxy_list import ProxyList
import numpy as np

try:
    from sys import maxint
except:
    from sys import maxsize as maxint


class SourceProxyList(ProxyList):

    def __init__(self, obj):
        super(SourceProxyList, self).__init__(obj, "_source_count",
                                              "_get_source_by_id",
                                              "_get_source_by_pos",
                                              "_delete_source_by_id")


class DataArrayProxyList(ProxyList):

    def __init__(self, obj):
        super(DataArrayProxyList, self).__init__(obj, "_data_array_count",
                                                 "_get_data_array_by_id",
                                                 "_get_data_array_by_pos",
                                                 "_delete_data_array_by_id")


class MultiTagProxyList(ProxyList):

    def __init__(self, obj):
        super(MultiTagProxyList, self).__init__(obj, "_multi_tag_count",
                                                "_get_multi_tag_by_id",
                                                "_get_multi_tag_by_pos",
                                                "_delete_multi_tag_by_id")


class TagProxyList(ProxyList):

    def __init__(self, obj):
        super(TagProxyList, self).__init__(obj, "_tag_count", "_get_tag_by_id",
                                           "_get_tag_by_pos",
                                           "_delete_tag_by_id")


class GroupProxyList(ProxyList):

    def __init__(self, obj):
        super(GroupProxyList, self).__init__(obj, "_group_count",
                                             "_get_group_by_id",
                                             "_get_group_by_pos",
                                             "_delete_group_by_id")


class BlockMixin(object):

    def create_data_array(self, name, array_type,
                          dtype=None, shape=None, data=None):
        """
        Create a new data array for this block. Either ``shape``
        or ``data`` must be given. If both are given their shape must agree.
        If ``dtype`` is not specified it will default to 64-bit floating
        points.

        :param name: The name of the data array to create.
        :type name: str
        :param array_type: The type of the data array.
        :type array_type: str
        :param dtype: Which data-type to use for storage
        :type dtype:  :class:`numpy.dtype`
        :param shape: Layout (dimensionality and extent)
        :type shape: tuple of int or long
        :param data: Data to write after storage has been created
        :type data: array-like data

        :returns: The newly created data array.
        :rtype: :class:`~nixio.DataArray`
        """

        if data is None:
            if shape is None:
                raise ValueError("Either shape and or data must not be None")
            if dtype is None:
                dtype = 'f8'
        else:
            data = np.ascontiguousarray(data)
            if dtype is None:
                dtype = data.dtype
            if shape is not None:
                if shape != data.shape:
                    raise ValueError("Shape must equal data.shape")
            else:
                shape = data.shape
        da = self._create_data_array(name, array_type, dtype, shape)
        if data is not None:
            da.write_direct(data)
        return da

    def find_sources(self, filtr=lambda _: True, limit=None):
        """
        Get all sources in this block recursively.

        This method traverses the tree of all sources in the block. The
        traversal is accomplished via breadth first and can be limited in
        depth.  On each node or source a filter is applied. If the filter
        returns true the respective source will be added to the result list.
        By default a filter is used that accepts all sources.

        :param filtr: A filter function
        :type filtr:  function
        :param limit: The maximum depth of traversal
        :type limit:  int

        :returns: A list containing the matching sources.
        :rtype: list of Source
        """
        if limit is None:
            limit = maxint
        return finders._find_sources(self, filtr, limit)

    @property
    def sources(self):
        """
        A property containing all sources of a block. Sources can be obtained
        via their index or by their id. Sources can be deleted from the list.
        Adding sources is done using the Blocks create_source method.
        This is a read only attribute.

        :type: ProxyList of Source entities.
        """
        if not hasattr(self, "_sources"):
            setattr(self, "_sources", SourceProxyList(self))
        return self._sources

    @property
    def multi_tags(self):
        """
        A property containing all multi tags of a block. MultiTag entities can
        be obtained via their index or by their id. Tags can be deleted from
        the list. Adding tags is done using the Blocks create_multi_tag method.
        This is a read only attribute.

        :type: ProxyList of MultiTag entities.
        """
        if not hasattr(self, "_multi_tags"):
            setattr(self, "_multi_tags", MultiTagProxyList(self))
        return self._multi_tags

    @property
    def tags(self):
        """
        A property containing all tags of a block. Tag entities can be obtained
        via their index or by their id. Tags can be deleted from the list.
        Adding tags is done using the Blocks create_tag method.
        This is a read only attribute.

        :type: ProxyList of Tag entities.
        """
        if not hasattr(self, "_tags"):
            setattr(self, "_tags", TagProxyList(self))
        return self._tags

    @property
    def data_arrays(self):
        """
        A property containing all data arrays of a block. DataArray entities
        can be obtained via their index or by their id. Data arrays can be
        deleted from the list. Adding a data array is done using the Blocks
        create_data_array method.
        This is a read only attribute.

        :type: ProxyList of DataArray entities.
        """
        if not hasattr(self, "_data_arrays"):
            setattr(self, "_data_arrays", DataArrayProxyList(self))
        return self._data_arrays

    @property
    def groups(self):
        """
        A property containing all groups of a block. Group entities can be
        obtained via their index or by their id. Groups can be deleted from the
        list. Adding a Group is done using the Blocks create_group method.
        This is a read only attribute.

        :type: ProxyList of Group entities.
        """
        if not hasattr(self, "_groups"):
            setattr(self, "_groups", GroupProxyList(self))
        return self._groups

    def __eq__(self, other):
        """
        Two blocks are considered equal when they have the same id.
        """
        if hasattr(other, "id"):
            return self.id == other.id
        else:
            return False

    def __hash__(self):
        """
        overwriting method __eq__ blocks inheritance of __hash__ in Python 3
        hash has to be either explicitly inherited from parent class,
        implemented or escaped
        """
        return hash(self.id)



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

import numpy as np

from ..value import DataType
from . import util


class H5DataSet(object):

    def __init__(self, parent, name, dtype=None, shape=None):
        self._parent = parent
        self.name = name
        if (dtype is None) or (shape is None):
            self.dataset = self._parent[name]
        else:
            maxshape = (None,) * len(shape)
            if dtype == DataType.String:
                dtype = util.vlen_str_dtype
            self.dataset = self._parent.require_dataset(name, shape=shape,
                                                        dtype=dtype,
                                                        chunks=True,
                                                        maxshape=maxshape)

    @classmethod
    def create_from_h5obj(cls, h5obj):
        parent = h5obj.parent
        name = h5obj.name.split("/")[-1]
        return cls(parent, name)

    def write_data(self, data, count=None, offset=None):
        if count and offset:
            sl = util.co_to_slice(count, offset)
            self.dataset[sl] = data
        else:
            self.dataset[:] = data

    def read_data(self, data, count=None, offset=None):
        if 0 in self.dataset.shape or len(self.dataset.shape) == 0:
            return
        if count and offset:
            if sum(count) == 0 and len(data) == 0:
                return data
            datashape = data.shape
            sl = util.co_to_slice(count, offset)
            if (isinstance(sl, tuple) and
                    np.ndim(data) != len(sl) and
                    count[-1] == 1):
                    # data needs an extra dimension of size 1 to match slice
                    data.resize(datashape + (1,))
            self.dataset.read_direct(data, sl)
            # if data was resized, it should be returned to its original shape
            data.resize(datashape)
        else:
            self.dataset.read_direct(data)

    def set_attr(self, name, value):
        if value is None:
            if name in self.dataset.attrs:
                del self.dataset.attrs[name]
        else:
            self.dataset.attrs[name] = value

    def get_attr(self, name):
        attr = self.dataset.attrs.get(name)
        if isinstance(attr, bytes):
            attr = attr.decode()
        return attr

    @property
    def shape(self):
        return self.dataset.shape

    @shape.setter
    def shape(self, shape):
        self.dataset.resize(shape)

    @property
    def dtype(self):
        return self.dataset.dtype

    def __str__(self):
        return "<H5DataSet object: {}>".format(self.dataset.name)



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from .entity import Entity
from .data_array import DataArray
from .util import util


class Feature(Entity):

    def __init__(self, nixparent, h5group):
        super(Feature, self).__init__(nixparent, h5group)

    @classmethod
    def _create_new(cls, nixparent, h5parent, data, link_type):
        newentity = super(Feature, cls)._create_new(nixparent, h5parent)
        newentity.link_type = link_type
        newentity.data = data
        return newentity

    @property
    def id(self):
        return self._h5group.get_attr("entity_id")

    @property
    def link_type(self):
        return util.link_type_from_string(self._h5group.get_attr("link_type"))

    @link_type.setter
    def link_type(self, lt):
        self._h5group.set_attr("link_type", util.link_type_to_string(lt))

    @property
    def data(self):
        return DataArray(self._parent._parent,
                         self._h5group.open_group("data"))

    @data.setter
    def data(self, da):
        if da is None:
            raise TypeError("Feature.data cannot be None.")
        parblock = self._parent._parent
        if da not in parblock.data_arrays:
            raise RuntimeError("Feature.data: DataArray not found in Block!")
        if "data" in self._h5group:
            del self._h5group["data"]
        self._h5group.create_link(da, "data")
        self.force_updated_at()



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division)
from numbers import Number

import numpy as np

from ..value import DataType
from ..dimension_type import DimensionType
from . import util


class Dimension(object):

    def __init__(self, h5group, index):
        from nixio.pycore.h5group import H5Group
        if not isinstance(h5group, H5Group):
            raise Exception
        self._h5group = h5group
        self.dim_index = index

    @classmethod
    def _create_new(cls, parent, index):
        h5group = parent.open_group(str(index))
        newdim = cls(h5group, index)
        return newdim

    @property
    def dimension_type(self):
        return self._h5group.get_attr("dimension_type")

    @dimension_type.setter
    def dimension_type(self, dimtype):
        if dimtype not in (DimensionType.Sample, DimensionType.Range,
                           DimensionType.Set):
            raise ValueError("Invalid dimension type.")
        self._h5group.set_attr("dimension_type", dimtype)

    @property
    def index(self):
        return self.dim_index

    @index.setter
    def index(self, idx):
        util.check_attr_type(idx, int)
        self.dim_index = idx


class SampledDimension(Dimension):

    def __init__(self, h5group, index):
        super(SampledDimension, self).__init__(h5group, index)

    @classmethod
    def _create_new(cls, parent, index, sample):
        newdim = super(SampledDimension, cls)._create_new(parent, index)
        newdim.dimension_type = DimensionType.Sample
        newdim.sampling_interval = sample
        return newdim

    def position_at(self, index):
        """
        Returns the position corresponding to a given index.

        :param index: A positive integer.

        :returns: The position matching to the index.
        :rtype: float
        """
        offset = self.offset if self.offset else 0
        sample = self.sampling_interval
        return index * sample + offset

    def index_of(self, position):
        """
        Returns the index of a certain position in the dimension.

        :param position: The position.

        :returns: The nearest index.
        :rtype: int
        """
        offset = self.offset if self.offset else 0
        sample = self.sampling_interval
        index = round((position - offset) / sample)
        if index < 0:
            raise IndexError("Position is out of bounds of this dimension!")
        return index

    def axis(self, count, start=0):
        """
        Get an axis as defined by this sampled dimension.

        :param count: A positive integer specifying the length of the axis
        (no of samples).
        :param start: positive integer, indicates the starting sample.

        :returns: The created axis
        :rtype: list
        """
        offset = self.offset if self.offset else 0
        sample = self.sampling_interval
        end = (count + start) * sample + offset
        return tuple(np.arange(offset, end, sample))

    @property
    def label(self):
        return self._h5group.get_attr("label")

    @label.setter
    def label(self, l):
        util.check_attr_type(l, str)
        self._h5group.set_attr("label", l)

    @property
    def sampling_interval(self):
        return self._h5group.get_attr("sampling_interval")

    @sampling_interval.setter
    def sampling_interval(self, interval):
        util.check_attr_type(interval, Number)
        self._h5group.set_attr("sampling_interval", interval)

    @property
    def unit(self):
        return self._h5group.get_attr("unit")

    @unit.setter
    def unit(self, u):
        util.check_attr_type(u, str)
        self._h5group.set_attr("unit", u)

    @property
    def offset(self):
        return self._h5group.get_attr("offset")

    @offset.setter
    def offset(self, o):
        util.check_attr_type(o, Number)
        self._h5group.set_attr("offset", o)


class RangeDimension(Dimension):

    def __init__(self, h5group, index):
        super(RangeDimension, self).__init__(h5group, index)

    @classmethod
    def _create_new(cls, parent, index, ticks):
        newdim = super(RangeDimension, cls)._create_new(parent, index)
        newdim.dimension_type = DimensionType.Range
        ticksds = newdim._h5group.create_dataset("ticks",
                                                 shape=np.shape(ticks),
                                                 dtype=DataType.Double)
        ticksds.write_data(ticks)
        return newdim

    @classmethod
    def _create_new_alias(cls, parent, index, da):
        newdim = super(RangeDimension, cls)._create_new(parent, index)
        newdim.dimension_type = DimensionType.Range
        newdim._h5group.create_link(da, da.id)
        return newdim

    @property
    def _alias(self):
        """
        Return True if this dimension is an Alias Range dimension.
        Read-only property.
        """
        if self._h5group.has_data("ticks"):
            return False
        return True

    @property
    def ticks(self):
        g = self._redirgrp
        if g.has_data("ticks"):
            ticks = g.get_data("ticks")
        elif g.has_data("data"):
            ticks = g.get_data("data")
        else:
            raise AttributeError("Attribute 'ticks' is not set.")
        return tuple(ticks)

    @ticks.setter
    def ticks(self, ticks):
        if np.any(np.diff(ticks) < 0):
            raise ValueError("Ticks are not given in an ascending order.")
        ticksds = self._h5group.get_dataset("ticks")
        ticksds.write_data(ticks)

    @property
    def _redirgrp(self):
        """
        If the dimension is an Alias Range dimension, this property returns
        the H5Group of the linked DataArray. Otherwise, it returns the H5Group
        representing the dimension.
        """
        if self._alias:
            gname = self._h5group.get_by_pos(0).name
            return self._h5group.open_group(gname)
        return self._h5group

    @property
    def label(self):
        return self._redirgrp.get_attr("label")

    @label.setter
    def label(self, l):
        util.check_attr_type(l, str)
        self._redirgrp.set_attr("label", l)

    @property
    def unit(self):
        return self._redirgrp.get_attr("unit")

    @unit.setter
    def unit(self, u):
        util.check_attr_type(u, str)
        self._redirgrp.set_attr("unit", u)

    def index_of(self, position):
        """
        Returns the index of a certain position in the dimension.

        :param position: The position.

        :returns: The nearest index.
        :rtype: int
        """
        ticks = self.ticks
        if position < ticks[0]:
            return 0
        elif position > ticks[-1]:
            return len(ticks) - 1

        ticks = np.array(ticks)
        pidxs = np.flatnonzero((ticks - position) >= 0)
        return pidxs[0]

    def tick_at(self, index):
        """
        Returns the tick at the given index. Will throw an Exception if the
        index is out of bounds.

        :param index: The index.

        :returns: The corresponding position.
        :rtype: double
        """
        ticks = list(self.ticks)
        return ticks[index]

    def axis(self, count, start=0):
        """
        Get an axis as defined by this range dimension.

        :param count: A positive integer specifying the length of the axis
        (no of points).
        :param start: positive integer, indicates the starting tick.

        :returns: The created axis
        :rtype: list
        """
        ticks = self.ticks
        end = start + count
        if end > len(ticks):
            raise IndexError("RangeDimension.axis: Count is invalid, "
                             "reaches beyond the ticks stored in this "
                             "dimension.")
        return ticks[start:end]


class SetDimension(Dimension):

    def __init__(self, h5group, index):
        super(SetDimension, self).__init__(h5group, index)

    @classmethod
    def _create_new(cls, parent, index):
        newdim = super(SetDimension, cls)._create_new(parent, index)
        newdim.dimension_type = DimensionType.Set
        return newdim

    @property
    def labels(self):
        labels = tuple(self._h5group.get_data("labels"))
        if len(labels) and isinstance(labels[0], bytes):
            labels = tuple(l.decode() for l in labels)
        return labels

    @labels.setter
    def labels(self, labels):
        lshape = np.shape(labels)
        dt = util.vlen_str_dtype
        labelsds = self._h5group.create_dataset("labels", shape=lshape,
                                                dtype=dt)
        labelsds.write_data(labels)



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)
import os
import gc
from warnings import warn
import numpy as np

import h5py

from .h5group import H5Group
from .block import Block
from .exceptions import exceptions
from .section import Section
from ..file import FileMixin
from . import util

try:
    from ..core import File as CFile
except ImportError:
    CFile = None


FILE_FORMAT = "nix"
HDF_FF_VERSION = (1, 1, 0)


def can_write(nixfile):
    filever = nixfile.version
    if len(filever) != 3:
        raise RuntimeError("Invalid version specified in file.")
    if HDF_FF_VERSION == filever:
        return True
    else:
        return False


def can_read(nixfile):
    filever = nixfile.version
    if len(filever) != 3:
        raise RuntimeError("Invalid version specified in file.")
    vx, vy, vz = HDF_FF_VERSION
    fx, fy, fz = filever
    if vx == fx and vy >= fy:
        return True
    else:
        return False


class FileMode(object):
    ReadOnly = 'r'
    ReadWrite = 'a'
    Overwrite = 'w'


def map_file_mode(mode):
    if mode == FileMode.ReadOnly:
        return h5py.h5f.ACC_RDONLY
    elif mode == FileMode.ReadWrite:
        return h5py.h5f.ACC_RDWR
    elif mode == FileMode.Overwrite:
        return h5py.h5f.ACC_TRUNC
    else:
        raise ValueError("Invalid file mode specified.")


def make_fapl():
    return h5py.h5p.create(h5py.h5p.FILE_ACCESS)


def make_fcpl():
    fcpl = h5py.h5p.create(h5py.h5p.FILE_CREATE)
    flags = h5py.h5p.CRT_ORDER_TRACKED | h5py.h5p.CRT_ORDER_INDEXED
    fcpl.set_link_creation_order(flags)
    return fcpl


class File(FileMixin):

    def __init__(self, h5file):
        self._h5file = h5file
        self._root = H5Group(self._h5file, "/", create=True)
        self._h5group = self._root  # to match behaviour of other objects
        self._data = self._root.open_group("data", create=True)
        self.metadata = self._root.open_group("metadata", create=True)
        if "created_at" not in self._h5file.attrs:
            self.force_created_at()
        if "updated_at" not in self._h5file.attrs:
            self.force_updated_at()

    @classmethod
    def _open_existing(cls, path, h5mode):
        fid = h5py.h5f.open(path, flags=h5mode, fapl=make_fapl())
        h5file = h5py.File(fid)
        nixfile = cls(h5file)
        return nixfile

    @classmethod
    def _create_new(cls, path, h5mode):
        fid = h5py.h5f.create(path, flags=h5mode, fapl=make_fapl(),
                              fcpl=make_fcpl())
        h5file = h5py.File(fid)
        newfile = cls(h5file)
        newfile._create_header()
        return newfile

    @classmethod
    def _open(cls, path, mode=FileMode.ReadWrite):
        try:
            path = path.encode("utf-8")
        except (UnicodeError, LookupError):
            pass

        if not os.path.exists(path) and mode == FileMode.ReadOnly:
            raise RuntimeError(
                "Cannot open non-existent file in ReadOnly mode!"
            )

        if not os.path.exists(path) or mode == FileMode.Overwrite:
            mode = FileMode.Overwrite
            h5mode = map_file_mode(mode)
            newfile = cls._create_new(path, h5mode)
            newfile.mode = mode
            return newfile

        h5mode = map_file_mode(mode)
        newfile = cls._open_existing(path, h5mode)
        newfile._check_header(mode)
        newfile.mode = mode
        return newfile

    @classmethod
    def open(cls, path, mode=FileMode.ReadWrite, backend=None):
        """
        Open a NIX file, or create it if it does not exist.

        :param path: Path to file
        :param mode: FileMode ReadOnly, ReadWrite, or Overwrite.
                    (default: ReadWrite)
        :param backend: Either "hdf5" or "h5py".
                        Defaults to "hdf5" if available, or "h5py" otherwise
        :return: nixio.File object
        """
        if backend is None:
            backend = os.getenv("NIXPY_H5_BACKEND")
        if backend is None:
            if CFile is None:
                backend = "h5py"
            else:
                backend = "hdf5"
        if backend == "hdf5":
            if CFile:
                return CFile.open(path, mode)
            else:
                raise RuntimeError("HDF5 backend is not available.")
        elif backend == "h5py":
            return cls._open(path, mode)
        else:
            raise ValueError("Valid backends are 'hdf5' and 'h5py'.")

    def _create_header(self):
        self.format = FILE_FORMAT
        self.version = HDF_FF_VERSION

    def _check_header(self, mode):
        if self.format != FILE_FORMAT:
            raise exceptions.InvalidFile

        if mode == FileMode.ReadWrite:
            if not can_write(self):
                raise RuntimeError("Cannot open file for writing. "
                                   "Incompatible version.")
        elif mode == FileMode.ReadOnly:
            if not can_read(self):
                raise RuntimeError("Cannot open file. "
                                   "Incompatible version.")

    @property
    def version(self):
        """
        The file format version.

        :type: tuple
        """
        return tuple(self._root.get_attr("version"))

    @version.setter
    def version(self, v):
        util.check_attr_type(v, tuple)
        for part in v:
            util.check_attr_type(part, int)
        # convert to np.int32 since py3 defaults to 64
        v = np.array(v, dtype=np.int32)
        self._root.set_attr("version", v)

    @property
    def format(self):
        """
        The format of the file. This read only property should always have the
        value 'nix'.

        :type: str
        """
        return self._root.get_attr("format")

    @format.setter
    def format(self, f):
        util.check_attr_type(f, str)
        self._root.set_attr("format", f.encode("ascii"))

    @property
    def created_at(self):
        """
        The creation time of the file. This is a read-only property.
        Use `force_created_at` in order to change the creation time.

        :rtype: int
        """
        return util.str_to_time(self._h5file.attrs["created_at"])

    def force_created_at(self, t=None):
        """
        Sets the creation time `created_at` to the given time
        (default: current time).

        :param t: The time to set
        :type t: int
        """
        if t is None:
            t = util.now_int()
        else:
            util.check_attr_type(t, int)
        self._h5file.attrs["created_at"] = util.time_to_str(t)

    @property
    def updated_at(self):
        """
        The time of the last update of the file. This is a read-only
        property. Use `force_updated_at` in order to change the update
        time.

        :rtype: int
        """
        return util.str_to_time(self._h5file.attrs["updated_at"])

    def force_updated_at(self, t=None):
        """
        Sets the update time `updated_at` to the given time.
        (default: current time)

        :param t: The time to set (default: now)
        :type t: int
        """
        if t is None:
            t = util.now_int()
        else:
            util.check_attr_type(t, int)
        self._h5file.attrs["updated_at"] = util.time_to_str(t)

    def validate(self):
        """
        Checks if the File is a valid NIX file. This method is only available
        when using the "hdf5" backend.

        :return: Result object
        """
        warn("The h5py backend does not support validation.")

    def is_open(self):
        """
        Checks whether a file is open or closed.

        :returns: True if the file is open, False otherwise.
        :rtype: bool
        """
        try:
            self._h5file.mode
            return True
        except ValueError:
            return False

    def flush(self):
        self._h5file.flush()

    def close(self):
        """
        Closes an open file.
        """
        gc.collect()  # should handle refs better instead of calling collect()
        # Flush is probably unnecessary
        self._h5file.flush()
        self._h5file.close()

    # Block
    def create_block(self, name, type_):
        """
        Create a new block inside the file.

        :param name: The name of the block to create.
        :type name: str
        :param type_: The type of the block.
        :type type_: str

        :returns: The newly created block.
        :rtype: Block
        """
        if name in self._data:
            raise ValueError("Block with the given name already exists!")
        block = Block._create_new(self, self._data, name, type_)
        return block

    def _get_block_by_id(self, id_or_name):
        return Block(self, self._data.get_by_id_or_name(id_or_name))

    def _get_block_by_pos(self, pos):
        return Block(self, self._data.get_by_pos(pos))

    def _delete_block_by_id(self, id_):
        self._data.delete(id_)

    def _block_count(self):
        return len(self._data)

    # Section
    def create_section(self, name, type_):
        """
        Create a new metadata section inside the file.

        :param name: The name of the section to create.
        :type name: str
        :param type_: The type of the section.
        :type type_: str

        :returns: The newly created section.
        :rtype: Section
        """
        if name in self.metadata:
            raise exceptions.DuplicateName("create_section")
        sec = Section._create_new(self, self.metadata, name, type_)
        return sec

    def _get_section_by_id(self, id_or_name):
        return Section(self, self.metadata.get_by_id_or_name(id_or_name))

    def _get_section_by_pos(self, pos):
        return Section(self, self.metadata.get_by_pos(pos))

    def _delete_section_by_id(self, id_):
        self.metadata.delete(id_)

    def _section_count(self):
        return len(self.metadata)



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
import numpy as np

from .data_array import DataSet
from .exceptions import OutOfBounds


class DataView(DataSet):

    def __init__(self, da, count, offset):
        self.array = da
        self._h5group = self.array._h5group
        self._count = tuple(count)
        self._offset = tuple(offset)

        co = tuple(c + o for c, o in zip(self._count, self._offset))
        if any(c > e for c, e in zip(co, self.array.data_extent)):
            raise OutOfBounds(
                "Trying to create DataView which is out of bounds"
            )

    @property
    def data_extent(self):
        return self._count

    @property
    def data_type(self):
        return self.array.data_type

    def _write_data(self, data, count, offset):
        if not count:
            count = self._count
        offset = self._transform_coordinates(count, offset)
        return super(DataView, self)._write_data(data, count, offset)

    def _read_data(self, data, count, offset):
        if not count:
            count = self._count
        offset = self._transform_coordinates(count, offset)
        return super(DataView, self)._read_data(data, count, offset)

    def _transform_coordinates(self, count, offset):
        if not offset:
            if np.any(np.greater(count, self._count)):
                raise OutOfBounds("Trying to access data outside of range")
            return self._offset
        else:
            co = tuple(c + o for c, o in zip(count, offset))
            if any(c > sc for c, sc in zip(co, self._count)):
                raise OutOfBounds("Trying to access data outside of range")
            return tuple(so + o for so, o in zip(self._offset, offset))



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

from .entity_with_metadata import EntityWithMetadata
from ..entity_with_sources import EntityWithSourcesMixin
from .source import Source

from . import util


class EntityWithSources(EntityWithMetadata, EntityWithSourcesMixin):

    def __init__(self, nixparent, h5group):
        super(EntityWithSources, self).__init__(nixparent, h5group)

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        newentity = super(EntityWithSources, cls)._create_new(
            nixparent, h5parent, name, type_
        )
        return newentity

    # Source
    def _get_source_by_id(self, id_or_name):
        sources = self._h5group.open_group("sources")
        if util.is_uuid(id_or_name):
            id_ = id_or_name
        else:
            for grp in sources:
                if grp.get_attr("name") == id_or_name:
                    id_ = grp.get_attr("entity_id")
                    break
            else:
                raise ValueError("No Source with name {} found {}.sources"
                                 .format(id_or_name, self.name))
        return Source(self, sources.get_by_name(id_))

    def _get_source_by_pos(self, pos):
        sources = self._h5group.open_group("sources")
        return Source(self, sources.get_by_pos(pos))

    def _remove_source_by_id(self, id_):
        sources = self._h5group.open_group("sources")
        sources.delete(id_)

    def _source_count(self):
        sources = self._h5group.open_group("sources")
        return len(sources)

    def _add_source_by_id(self, id_):
        parblock = self._parent
        target = parblock._h5group.find_children(
            filtr=lambda x: x.get_attr("entity_id") == id_
        )
        cls = type(self).__name__
        if not target:
            raise RuntimeError("{}._add_source_by_id: "
                               "Source not found!".format(cls))
        if len(target) > 1:
            raise RuntimeError("{}._add_source_by_id: "
                               "Invalid data found in NIX file. "
                               "Multiple Sources found with the same ID."
                               .format(cls))
        target = Source(parblock, target[0])
        sources = self._h5group.open_group("sources")
        sources.create_link(target, target.id)

    def _has_source_by_id(self, id_or_name):
        sources = self._h5group.open_group("sources")
        sources.has_by_id(id_or_name)



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function)

import sys
import os

_nixio_bin = os.path.join(sys.prefix, 'share', 'nixio', 'bin')
if os.path.isdir(_nixio_bin):
    os.environ["PATH"] += os.pathsep + _nixio_bin

from nixio.pycore.file import File, FileMode
from nixio.value import Value, DataType
from nixio.dimension_type import DimensionType
from nixio.link_type import LinkType

from nixio.section import S

from nixio.info import VERSION as __version__

try:
    import nixio.util.inject
except ImportError:
    pass

__all__ = ("File", "FileMode", "DataType", "Value",
           "LinkType", "DimensionType")
__author__ = ('Christian Kellner, Adrian Stoewer, Andrey Sobolev, Jan Grewe, '
              'Balint Morvai, Achilleas Koutsou')



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

import nixio


class Cont(object):
    """
    Simple container for an element an a level
    """

    def __init__(self, elem, level):
        self.elem = elem
        self.level = level


def _find_sources(with_sources, filtr, limit):
    """
    Find a list of matching sources recursively.
    For internal use.
    """

    fifo = []
    result = []
    level = 0

    sourcecls = (nixio.pycore.Source,)
    try:
        sourcecls += (nixio.core.Source,)
    except AttributeError:
        pass

    if isinstance(with_sources, sourcecls):
        fifo.append(Cont(with_sources, level))
    else:
        level += 1
        fifo += [Cont(e, level) for e in with_sources.sources]

    while len(fifo) > 0:
        c = fifo.pop(0)

        level = c.level + 1
        if level <= limit:
            fifo += [Cont(e, level) for e in c.elem.sources]

        if filtr(c.elem):
            result.append(c.elem)

    return result


def _find_sections(with_sections, filtr, limit):
    """
    Find a list of matching sections recursively.
    For internal use.
    """

    fifo = []
    result = []
    level = 0

    sectioncls = (nixio.pycore.Section,)
    try:
        sectioncls += (nixio.core.Section,)
    except AttributeError:
        pass

    if isinstance(with_sections, sectioncls):
        fifo.append(Cont(with_sections, level))
    else:
        level += 1
        fifo += [Cont(e, level) for e in with_sections.sections]

    while len(fifo) > 0:
        c = fifo.pop(0)

        level = c.level + 1
        if level <= limit:
            fifo += [Cont(e, level) for e in c.elem.sections]

        if filtr(c.elem):
            result.append(c.elem)

    return result



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

VERSION = '1.4.1'
STATUS = 'Release'
RELEASE = '%s %s' % (VERSION, STATUS)
AUTHOR = 'Christian Kellner, Adrian Stoewer, Andrey Sobolev, Jan Grewe,\
Balint Morvai, Achilleas Koutsou'
COPYRIGHT = '2014, German Neuroinformatics Node, ' + AUTHOR
CONTACT = 'dev@g-node.org'
BRIEF = 'Python bindings for NIX'
HOMEPAGE = 'https://github.com/G-Node/nixpy'
